Imports System.Net
Imports System.Net.Sockets
Imports System.IO
Imports TcpCommon
Imports System.Drawing.Imaging
Imports System.IO.Compression

Public Class frmClient

   Private WithEvents _EndPoint As TcpEndpoint
   Private _TextWriter As StreamWriter, _BinWriter As BinaryWriter

   Public Sub New()
      Dim bytes = BitConverter.GetBytes(Not Short.MaxValue)
      InitializeComponent()
      With Application.OpenForms 'bei mehreren Instanzen versetzt anordnen
         If .Count > 0 Then
            Me.Location = .Item(.Count - 1).Location + New Size(30, 50)
         Else
            Me.Location = Screen.PrimaryScreen.WorkingArea.Location
         End If
      End With
   End Sub

   Private Sub frmClient_Shown(sender As Object, e As EventArgs) Handles Me.Shown
      'fr schnelleren Startup beim Debuggen
      BeginInvoke(Sub()
                     btConnect.PerformClick()
                     ActionToolStripMenuItem.ShowDropDown()
                  End Sub)
   End Sub

   Dim zipper As GZipStream
   Private Sub GzipTest1()
      'zeigt, dass ein GZipStream bei Flush nur einen Teil wegschreibt, und erst beim 2. Klicks, bei Dispose schreibt er den Rest
      _EndPoint.WriteCommandCode(ServerCommand.Test1)
      If zipper Is Nothing Then
         zipper = New GZipStream(_EndPoint.WriteStream, CompressionMode.Compress)
         Dim wr = New StreamWriter(zipper)
         wr.WriteLine(rtbSend.Text)
         wr.Flush()
         zipper.Flush() 'schickt nur ein Teil
      Else
         zipper.Dispose() 'schickt den Rest
         zipper = Nothing
      End If
   End Sub

   Private Sub GzipTest2()
      'zeigt, dass ein GZipStream-Empfnger nur die erste GZip-Nachricht entpackt - die Bytes der 2. gehen im Lese-Puffer verloren.
      _EndPoint.WriteCommandCode(ServerCommand.Test2)
      Using zipper = New GZipStream(_EndPoint.WriteStream, CompressionMode.Compress), wr = New StreamWriter(zipper)
         wr.Write("Hallo") 'schickt 5 bytes
      End Using
      Using zipper = New GZipStream(_EndPoint.WriteStream, CompressionMode.Compress), wr = New StreamWriter(zipper)
         wr.Write("Hallo") 'schickt 5 bytes
      End Using
   End Sub

   Private Sub MenuItem_Click(ByVal sender As Object, ByVal e As EventArgs) Handles btConnect.Click, btTest1.Click, btOuterIP.Click, btAddClient.Click, btBroadcastText.Click, btBroadcastImage.Click, btInvertText.Click, btInvertNumber.Click, btInvertImage.Click, btSelectImage.Click, btSendFile.Click, btTest2.Click
      Select Case True
         Case sender Is btConnect : Connect()

         Case sender Is btOuterIP : frmOuterIP.Show(Me)

         Case sender Is btAddClient : Call (New frmClient).Show()

         Case sender Is btSelectImage 'zum testen auch mal sehr groe/kleine Bilder nehmen
            If OpenFileDialog1.ShowDialog(Me) <> Windows.Forms.DialogResult.OK Then Return
            PictureBox1.ImageLocation = OpenFileDialog1.FileName

         Case Else
            Select Case True 'server-communication
               Case sender Is btBroadcastImage
                  _EndPoint.WriteCommandCode(ServerCommand.ImageBroadcast)
                  SendImage()

               Case sender Is btInvertImage
                  _EndPoint.WriteCommandCode(ServerCommand.ImageInvert)
                  SendImage()

               Case sender Is btBroadcastText
                  _EndPoint.WriteCommandCode(ServerCommand.TextBroadcast)
                  _TextWriter.Write(Me.rtbSend.Text)
                  _TextWriter.Flush()

               Case sender Is btInvertText
                  _EndPoint.WriteCommandCode(ServerCommand.TextInvert)
                  _TextWriter.Write(Me.rtbSend.Text)
                  _TextWriter.Flush()     'StreamWriter needs flushing by Default - or set its AutoFlush-Property to True

               Case sender Is btInvertNumber
                  _EndPoint.WriteCommandCode(ServerCommand.NumberInvert)
                  _BinWriter.Write(NumericUpDown1.Value)
                  '_BinWriter.Flush()  'BinaryWriter needs no flushing

               Case sender Is btSendFile : SendCompressedFile()

               Case sender Is btTest1 : GzipTest1()

               Case sender Is btTest2 : GzipTest2()

            End Select
            _EndPoint.WriteStream.NextContent()
      End Select
   End Sub

   Private Sub Connect()
      If _EndPoint.NotNull Then _EndPoint.Dispose()
      Dim EP As IPEndPoint = GetIpEndPoint()
      If EP Is Nothing Then Return
      'Sehr rgerlich: der TcpClient-Konstruktor New(EP as IPEndPoint)
      'Produziert auf Einzelplatzsystemen einen unntigen Fehler, weil er (undokumentiert!) TcpClient.ExclusiveAddressUse=True voreinstellt. 
      Dim TC As New TcpClient
      TC.ExclusiveAddressUse = False
      Try
         TC.Connect(EP)
      Catch ex As SocketException
         TC.Close()
         DisplayStatusLine("Fehler:", Lf, ex.Message)
         Return
      End Try
      _EndPoint = New TcpEndpoint(TC)
      DisplayStatusLine("Hi!", Lf, "I'm a Client, connected with:", Lf, EP)
      _TextWriter = New StreamWriter(_EndPoint.WriteStream)
      _BinWriter = New BinaryWriter(_EndPoint.WriteStream)
   End Sub

   Private Sub SendImage()
      Dim img = PictureBox1.Image
      Using ms As New MemoryStream
         'Image.Save() bentigt einen Seekeablen Stream - daher in MemoryStream zwischenspeichern
         img.Save(ms, ImageFormat.Jpeg)
         ms.WriteTo(_EndPoint.WriteStream)
      End Using
   End Sub

   Private Sub SendCompressedFile()
      _EndPoint.WriteCommandCode(ServerCommand.ChangeCompressedFile)
      Using file = New FileStream(Path.GetFullPath("..\..\frmClient.vb"), FileMode.Open), zipper = New GZipStream(_EndPoint.WriteStream, CompressionLevel.Optimal)
         file.CopyTo(zipper)
      End Using
   End Sub

   'diese Sub ist normalerweise in der Designer.vb. Ich will aber sicherstellen kann, da _Endpoint **vor** den components disposed. Sonst empfngter womglich noch Daten, und will die darstellen.
   Protected Overrides Sub Dispose(ByVal disposing As Boolean)
      If disposing Then
         If _EndPoint.NotNull Then Dim ep = _EndPoint : _EndPoint = Nothing : ep.Dispose()
         If components IsNot Nothing Then components.Dispose()
      End If
      MyBase.Dispose(disposing)
   End Sub

   Private Sub Endpoint_DataReceived(ByVal sender As Object, ByVal e As EventArgs) Handles _EndPoint.DataReceived
      Invoke(DirectCast(AddressOf ProcessCommand, Action(Of TcpEndpoint)), sender)
   End Sub

   Private Sub _EndPoint_Disposed(sender As Object, e As EventArgs) Handles _EndPoint.Disposed
      BeginInvoke(Sub()    SetOffline("connection broken"))
   End Sub
   Private Sub SetOffline(msg As String)
      DisplayStatusLine(msg)
      _EndPoint = Nothing
      Me.Text = "Offline"
   End Sub

   Private Sub _Endpoint_Initialized(sender As Object, e As EventArgs) Handles _EndPoint.Initialized
      BeginInvoke(Sub() Me.Text = "Client-" & DirectCast(sender, TcpEndpoint).ID)
   End Sub

   Private Sub ProcessCommand(ep As TcpEndpoint)
      'von den 4 vorgesehenen ClientCommands sind immerhin 3 implementiert: Die Daten werden angezeigt - je nach ihrer Art
      Dim strm = ep.ReadStream
      Select Case DirectCast(ep.CommandCode, ClientCommand)
         Case ClientCommand.Number
            Using rd = New BinaryReader(strm)
               NumericUpDown1.Value = rd.ReadDecimal
            End Using

         Case ClientCommand.Image
            PictureBox1.Image.Dispose()
            PictureBox1.Image = Bitmap.FromStream(strm)

         Case ClientCommand.Text
            Using rd = New StreamReader(strm)
               rtbLog.AppendText(rd.ReadToEnd & Lf)
            End Using

         Case ClientCommand.File
            Using unzipper = New GZipStream(strm, CompressionMode.Decompress), rd = New StreamReader(unzipper)
               rtbLog.Text = rd.ReadToEnd
            End Using

         Case Else : Throw New NotSupportedException
      End Select
   End Sub

   ''' <summary>Meldung hellgelb hinterlegt anzeigen</summary>
   Private Sub DisplayStatusLine(ByVal ParamArray Items() As Object)
      Me.rtbLog.SelectionBackColor = Color.FromArgb(255, 255, 200)
      rtbLog.AppendText(String.Concat(Items) & Lf)
      Me.rtbLog.SelectionBackColor = Color.White
   End Sub
End Class