﻿Imports System.Net
Imports System.Net.Sockets
Imports TcpCommon
Imports System.IO
Imports System.Drawing
Imports System.Drawing.Imaging
Imports System.Threading
Imports System.IO.Compression
Imports System.Threading.Tasks

Module modServer

   Sub Main()
      Dim ep = GetIpEndPoint()
      If ep.Null Then
         Dbg("no valid IpEndpoint")
      Else
         Try
            Connect(ep)
            Dbg("Server listening to: ", ep.ToString)
         Catch ex As Exception
            Dbg("Error on try listening to: ", ep.ToString, Lf, ex.Message)
         End Try
      End If
      Console.ReadLine()
   End Sub

   Private _Listener As TcpListener

   'Pro Verbindung(-sanfrage) wird ein Client-Objekt generiert, das den Datenaustausch dieser Verbindung abwickelt
   Private _Clients As New List(Of TcpEndpoint)

   Private Async Sub Connect(ByVal EP As IPEndPoint)
      _Listener = New TcpListener(EP)
      _Listener.ExclusiveAddressUse = False
      _Listener.Start()
      For id = 0 To Integer.MaxValue
         Dim clnt = New TcpEndpoint(Await _Listener.AcceptTcpClientAsync, id)
         AddHandler clnt.DataReceived, AddressOf Client_DataReceived
         AddHandler clnt.Disposed, AddressOf Client_Disposed
         _Clients.Add(clnt)
         Dbg("Client-" & clnt.ID, "added")
         clnt.WriteStream.Write(BitConverter.GetBytes(id), 0, 4)
         clnt.WriteStream.NextContent()
      Next
   End Sub

   Private Sub Client_Disposed(ByVal sender As Object, ByVal e As EventArgs)
      Dim cl = DirectCast(sender, TcpEndpoint)
      _Clients.Remove(cl)
      Dbg("Client-" & cl.ID, "removed")
   End Sub

   Private Sub Client_DataReceived(ByVal sender As Object, ByVal e As EventArgs)
      Dim client = DirectCast(sender, TcpEndpoint)
      Dim cmd = DirectCast(client.CommandCode, ServerCommand)
      Dbg(String.Concat("Client-", client.ID, ": ", cmd))
      Select Case cmd 'BroadCast - Commands
         Case ServerCommand.TextBroadcast : TextBroadcast(client)
         Case ServerCommand.ImageBroadcast : ImageBroadcast(client)
         Case Else
            Select Case cmd ' Respond-to-Client - Commands
               Case ServerCommand.TextInvert : TextInvert(client)
               Case ServerCommand.NumberInvert : NumberInvert(client)
               Case ServerCommand.ChangeCompressedFile : ChangeCompressedFile(client)
               Case ServerCommand.ImageInvert : ImageInvert(client)
               Case ServerCommand.Test1
                  Using ms = New MemoryStream
                     client.ReadStream.CopyTo(ms)
                     Dbg("Anzahl empfangener Bytes:", ms.Length)
                  End Using
               Case ServerCommand.Test2
                  Using unzipper = New GZipStream(client.ReadStream, CompressionMode.Decompress, True), rd = New StreamReader(unzipper)
                     Dbg(rd.ReadToEnd)
                     Using rd2 = New StreamReader(unzipper)
                        Dbg(rd2.ReadToEnd)
                     End Using
                  End Using
                  Using ms = New MemoryStream
                     client.ReadStream.CopyTo(ms)
                     Dbg("Anzahl verbliebener Bytes nach Unzippen des ersten Contents:", ms.Length)
                  End Using

               Case Else : Dbg("unknown ServerCommand")
            End Select
            client.WriteStream.NextContent()
      End Select
   End Sub

   Private Sub ChangeCompressedFile(client As TcpEndpoint)
      client.WriteCommandCode(ClientCommand.File)
      Using unzipper = New GZipStream(client.ReadStream, CompressionMode.Decompress), zipper = New GZipStream(client.WriteStream, CompressionLevel.Optimal)
         zipper.WriteAllText(unzipper.ReadAllText.ToUpper)
      End Using
   End Sub

   Private Sub Client_StatusMessage(ByVal sender As Object, ByVal e As NotifyEventArgs(Of String))
      Dbg("(", e.Value, ")")
   End Sub

   Private Sub NumberInvert(client As TcpEndpoint)
      client.WriteCommandCode(ClientCommand.Number)
      Using rd = New BinaryReader(client.ReadStream), wr = New BinaryWriter(client.WriteStream)
         wr.Write(-rd.ReadDecimal)
      End Using
   End Sub
   Private Sub TextBroadcast(client As TcpEndpoint)
      Dim txt = client.ReadStream.ReadAllText
      Dbg(txt)
      For Each cl As TcpEndpoint In _Clients    ' an alle versenden
         cl.WriteCommandCode(ClientCommand.Text)
         cl.WriteStream.WriteAllText(txt)
         cl.WriteStream.NextContent()
      Next
   End Sub
   Private Sub TextInvert(client As TcpEndpoint)
      client.WriteCommandCode(ClientCommand.Text)
      client.WriteStream.WriteAllText(New String(client.ReadStream.ReadAllText.Reverse.ToArray))
   End Sub

   Private Sub ImageBroadcast(client As TcpEndpoint)
      Using ms As New MemoryStream
         client.ReadStream.WriteTo(ms)
         For Each cl In _Clients
            cl.WriteCommandCode(ClientCommand.Image)
            ms.WriteTo(cl.WriteStream)
            cl.WriteStream.NextContent()
         Next
      End Using
   End Sub

   Private Sub ImageInvert(client As TcpEndpoint)
      Dim img = Image.FromStream(client.ReadStream)
      img.RotateFlip(RotateFlipType.Rotate180FlipY)
      client.WriteCommandCode(ClientCommand.Image)
      img.Save(client.WriteStream, ImageFormat.Jpeg)
      Using ms As New MemoryStream
         img.Save(ms, ImageFormat.Jpeg)
         ms.WriteTo(client.WriteStream)
      End Using
   End Sub

End Module
